@@ -137,6 +137,8 @@ httpClientAbort(HTTPConnectionPtr connection, int closed)
     }
 }
 
+static void httpClientDelayedFinish(HTTPConnectionPtr connection);
+
 /* s is 1 in order to linger the connection, 2 to close it straight away. */
 void
 httpClientFinish(HTTPConnectionPtr connection, int s)
@@ -155,11 +157,25 @@ httpClientFinish(HTTPConnectionPtr connection, int s)
         dispose_chunk(connection->buf);
     connection->buf = NULL;
 
+    connection->flags &= ~CONN_WRITER;
+
+    if(connection->flags & CONN_SIDE_READER) {
+        /* We're in POST or PUT and the reader isn't done yet.
+           We make sure the request is marked as failed, and
+           reschedule ourselves. */
+        assert(request && (connection->flags & CONN_READER));
+        if(request->error_code == 0) {
+            request->error_code = 500;
+            request->error_message = internAtom("Error message lost");
+        }
+        httpClientDelayedFinish(connection);
+        return;
+    }
+
     if(connection->timeout) 
         cancelTimeEvent(connection->timeout);
     connection->timeout = NULL;
 
-    connection->flags &= ~CONN_WRITER;
     if(request) {
         HTTPRequestPtr requestee;
 
@@ -237,7 +253,7 @@ httpClientFinish(HTTPConnectionPtr connection, int s)
         if(request->ohandler)
             abortObjectHandler(request->ohandler);
         request->ohandler = NULL;
-        if(request->object->requestor == request)
+        if(request->object && request->object->requestor == request)
             request->object->requestor = NULL;
         httpDequeueRequest(connection);
         httpDestroyRequest(request);
@@ -269,6 +285,35 @@ httpClientFinish(HTTPConnectionPtr connection, int s)
     free(connection);
 }
 
+static int
+httpClientDelayedFinishHandler(TimeEventHandlerPtr event)
+{
+    HTTPConnectionPtr connection = *(HTTPConnectionPtr*)event->data;
+    httpClientFinish(connection, 1);
+    return 1;
+}
+
+static void
+httpClientDelayedFinish(HTTPConnectionPtr connection)
+{
+    TimeEventHandlerPtr handler;
+
+    handler = scheduleTimeEvent(1, httpClientDelayedFinishHandler,
+                                sizeof(connection), &connection);
+    if(!handler) {
+        do_log(L_ERROR, 
+               "Couldn't schedule delayed finish -- freeing memory.");
+        free_chunk_arenas();
+        handler = scheduleTimeEvent(1, httpClientDelayedFinishHandler,
+                                    sizeof(connection), &connection);
+        if(!handler) {
+            do_log(L_ERROR, 
+                   "Couldn't schedule delayed finish -- aborting.\n");
+            polipoExit();
+        }
+    }
+}
+
 /* Extremely baroque implementation of close: we need to synchronise
    between the writer and the reader.  */
 
@@ -428,10 +473,12 @@ httpClientRawErrorHeaders(HTTPConnectionPtr connection,
     assert(connection->flags & CONN_WRITER);
     assert(code != 0);
 
-    if(connection->request) 
-        close = close || !connection->request->persistent;
-    else
-        close = 1;
+    if(close >= 0) {
+        if(connection->request)
+            close = close || !connection->request->persistent;
+        else
+            close = 1;
+    }
 
     if(connection->request && connection->request->object) {
         url = connection->request->object->key;
@@ -454,21 +501,21 @@ httpClientRawErrorHeaders(HTTPConnectionPtr connection,
     n = httpWriteErrorHeaders(connection->buf, CHUNK_SIZE, 0,
                               connection->request &&
                               connection->request->method != METHOD_HEAD,
-                              code, message, close, headers,
+                              code, message, close > 0, headers,
                               url, url_len, etag);
     if(n <= 0) {
-        httpClientFinish(connection, 1);
+        shutdown(connection->fd, 1);
+        if(close >= 0)
+            httpClientFinish(connection, 1);
         return 1;
     }
 
     httpSetTimeout(connection, 60);
-    if(close) {
-        do_stream(IO_WRITE, fd, 0, connection->buf, n, 
-                  httpErrorStreamHandler, connection);
-    } else {
-        do_stream(IO_WRITE, fd, 0, connection->buf, n, 
-                  httpErrorNocloseStreamHandler, connection);
-    }
+    do_stream(IO_WRITE, fd, 0, connection->buf, n, 
+              close > 0 ? httpErrorStreamHandler :
+              close == 0 ? httpErrorNocloseStreamHandler :
+              httpErrorNofinishStreamHandler,
+              connection);
 
     return 1;
 }
@@ -577,6 +624,17 @@ httpErrorNocloseStreamHandler(int status,
     return 1;
 }
 
+int
+httpErrorNofinishStreamHandler(int status,
+                               FdEventHandlerPtr event,
+                               StreamRequestPtr srequest)
+{
+    if(status == 0 && !streamRequestDone(srequest))
+        return 0;
+
+    return 1;
+}
+
 int
 httpClientHandlerHeaders(FdEventHandlerPtr event, StreamRequestPtr srequest,
                          HTTPConnectionPtr connection)
@@ -756,13 +814,6 @@ httpClientRequest(HTTPRequestPtr request, AtomPtr url)
     if(expect) {
         if(expect == atom100Continue) {
             request->wait_continue = 1;
-        } else {
-            httpClientDiscardBody(connection);
-            httpClientNoticeError(request, 417,
-                                  internAtomF("Expectation %s not supported",
-                                              expect->string));
-            releaseAtom(expect);
-            return 1;
         }
         releaseAtom(expect);
     }
@@ -869,16 +920,19 @@ httpClientRequest(HTTPRequestPtr request, AtomPtr url)
     return 1;
 }
 
+static int httpClientDelayed(TimeEventHandlerPtr handler);
+
 int
 httpClientDiscardBody(HTTPConnectionPtr connection)
 {
-    assert(connection->request);
+    TimeEventHandlerPtr handler;
+
     assert(connection->reqoffset == 0);
 
     if(connection->reqte != TE_IDENTITY)
         goto fail;
 
-    if(connection->request->persistent && connection->bodylen < 0)
+    if(connection->bodylen < 0)
         goto fail;
 
     if(connection->bodylen + connection->reqbegin < connection->reqlen) {
@@ -910,33 +964,50 @@ httpClientDiscardBody(HTTPConnectionPtr connection)
         connection->reqlen = 0;
         connection->reqbegin = 0;
     }
+
     httpSetTimeout(connection, 60);
-    /* As we only notice new requests after that, the IO_NOTNOW is
-       necessary in order to prevent starvation if a client is
-       pipelining large numbers of requests. */
-    if(connection->reqlen > 0) {
-        do_stream(IO_READ | IO_NOTNOW | IO_IMMEDIATE,
-                  connection->fd, connection->reqlen,
-                  connection->reqbuf, CHUNK_SIZE,
-                  httpClientHandler, connection);
-    } else {
-        if(connection->reqbuf)
-            dispose_chunk(connection->reqbuf);
-        connection->reqbuf = NULL;
-        do_stream_buf(IO_READ | IO_NOTNOW,
-                      connection->fd, 0,
-                      &connection->reqbuf, CHUNK_SIZE,
-                      httpClientHandler, connection);
+    /* We need to delay in order to make sure the previous request
+       gets queued on the server side.  IO_NOTNOW isn't strong enough
+       for that due to IO_IMMEDIATE. */
+    handler = scheduleTimeEvent(-1, httpClientDelayed,
+                                sizeof(connection), &connection);
+    if(handler == NULL) {
+        do_log(L_ERROR, "Couldn't schedule reading from client.");
+        goto fail;
     }
     return 1;
 
  fail:
-    shutdown(connection->fd, 0);
-    connection->request->persistent = 0;
     connection->flags &= ~CONN_READER;
+    shutdown(connection->fd, 0);
+    httpClientFinish(connection, 1);
     return 1;
 }
 
+static int
+httpClientDelayed(TimeEventHandlerPtr event)
+{
+     HTTPConnectionPtr connection = *(HTTPConnectionPtr*)event->data;
+
+     /* IO_NOTNOW is unfortunate, but needed to avoid starvation if a
+        client is pipelining a lot of requests. */
+     if(connection->reqlen > 0) {
+         do_stream(IO_READ | IO_IMMEDIATE | IO_NOTNOW,
+                   connection->fd, connection->reqlen,
+                   connection->reqbuf, CHUNK_SIZE,
+                   httpClientHandler, connection);
+     } else {
+         if(connection->reqbuf)
+             dispose_chunk(connection->reqbuf);
+         connection->reqbuf = NULL;
+         do_stream_buf(IO_READ | IO_NOTNOW,
+                       connection->fd, 0,
+                       &connection->reqbuf, CHUNK_SIZE,
+                       httpClientHandler, connection);
+     }
+     return 1;
+}
+
 int
 httpClientDiscardHandler(int status,
                          FdEventHandlerPtr event, StreamRequestPtr request)
@@ -944,8 +1015,8 @@ httpClientDiscardHandler(int status,
     HTTPConnectionPtr connection = request->data;
 
     assert(connection->flags & CONN_READER);
-    if(status < 0) {
-        if(status != -EPIPE)
+    if(status) {
+        if(status < 0 && status != -EPIPE)
             do_log_error(L_ERROR, -status, "Couldn't read from client");
         connection->bodylen = -1;
         return httpClientDiscardBody(connection);
@@ -974,6 +1045,8 @@ httpClientNoticeRequest(HTTPRequestPtr request, int novalidate)
     if(request->error_code) {
         if(request->force_error || REQUEST_SIDE(request) ||
            request->object == NULL ||
+           (request->object->flags & OBJECT_LOCAL) ||
+           (request->object->flags & OBJECT_ABORTED) ||
            (relaxTransparency < 1 && !proxyOffline)) {
             if(serveNow) {
                 connection->flags |= CONN_WRITER;
@@ -1058,14 +1131,19 @@ httpClientNoticeRequest(HTTPRequestPtr request, int novalidate)
 
     if(request->cache_control.flags & CACHE_ONLY_IF_CACHED) {
         validate = 0;
-        if(!haveData)
+        if(!haveData) {
+            if(serveNow)
                 return httpClientRawError(connection, 504,
                                           internAtom("Object not in cache"),
                                           0);
+            else
+                return 1;
+        }
     }
 
     if(!(request->object->flags & OBJECT_VALIDATING) &&
-       ((!validate && haveData) || (request->object->flags & OBJECT_FAILED))) {
+       ((!validate && haveData) ||
+        (request->object->flags & OBJECT_FAILED))) {
         if(serveNow) {
             connection->flags |= CONN_WRITER;
             lockChunk(request->object, request->from / CHUNK_SIZE);
@@ -1119,8 +1197,10 @@ httpClientNoticeRequest(HTTPRequestPtr request, int novalidate)
         request->ohandler = NULL;
         request->object->flags &= ~OBJECT_VALIDATING;
         request->object->flags |= OBJECT_FAILED;
-        return httpClientRawError(connection, 503,
-                                  internAtom("Couldn't schedule get"), 0);
+        if(request->error_message)
+            releaseAtom(request->error_message);
+        request->error_code = 503;
+        request->error_message = internAtom("Couldn't schedule get");
     }
     return 1;
 }
@@ -1144,6 +1224,28 @@ delayedHttpClientNoticeRequest(HTTPRequestPtr request)
     return 1;
 }
 
+int
+httpClientContinueDelayed(TimeEventHandlerPtr event)
+{
+    static char httpContinue[] = "HTTP/1.1 100 Continue\r\n\r\n";
+    HTTPConnectionPtr connection = *(HTTPConnectionPtr*)event->data;
+
+    do_stream(IO_WRITE, connection->fd, 0, httpContinue, 25,
+              httpErrorNofinishStreamHandler, connection);
+    return 1;
+}
+
+int
+delayedHttpClientContinue(HTTPConnectionPtr connection)
+{
+    TimeEventHandlerPtr event;
+    event = scheduleTimeEvent(-1, httpClientContinueDelayed,
+                              sizeof(connection), &connection);
+    if(!event)
+        return -1;
+    return 1;
+}
+
 int
 httpClientGetHandler(int status, ObjectHandlerPtr ohandler)
 {
@@ -1154,6 +1256,9 @@ httpClientGetHandler(int status, ObjectHandlerPtr ohandler)
 
     assert(request == connection->request);
 
+    if(request->request)
+        assert(request->object->flags & OBJECT_INPROGRESS);
+
     if(status < 0) {
         object->flags &= ~OBJECT_VALIDATING; /* for now */
         if(request->request && request->request->request == request)
@@ -1184,6 +1289,48 @@ httpClientGetHandler(int status, ObjectHandlerPtr ohandler)
         return 1;
     }
 
+    if(request->wait_continue) {
+        if(request->request && request->request->wait_continue == 0) {
+            request->wait_continue = 0;
+            delayedHttpClientContinue(connection);
+        }
+        return 0;
+    }
+
+    /* See httpServerHandlerHeaders */
+    if((object->flags & OBJECT_SUPERSEDED) &&
+       request->request && request->request->can_mutate) {
+        ObjectPtr new_object = retainObject(request->request->can_mutate);
+        if(object->requestor == request && new_object->requestor == NULL) {
+            object->requestor = NULL;
+            new_object->requestor = request;
+        }
+        releaseObject(object);
+        object = new_object;
+        request->object = new_object;
+        request->request->object = new_object;
+    }
+
+    if(object->requestor != request && !(object->flags & OBJECT_ABORTED)) {
+        /* Make sure we don't serve an object that is stale for us
+           unless we're the requestor. */
+        if((object->flags & (OBJECT_LINEAR | OBJECT_MUTATING)) ||
+           objectMustRevalidate(object, &request->cache_control)) {
+           if(object->flags & OBJECT_INPROGRESS)
+               return 0;
+           rc = delayedHttpClientNoticeRequest(request);
+           if(rc < 0) {
+               do_log(L_ERROR, "Couldn't schedule noticing of request.");
+               abortObject(object, 500,
+                           internAtom("Couldn't schedule "
+                                      "noticing of request"));
+           } else {
+               request->ohandler = NULL;
+               return 1;
+           }
+        }
+    }
+
     if(object->flags & (OBJECT_INITIAL | OBJECT_VALIDATING)) {
         if(object->flags & (OBJECT_INPROGRESS | OBJECT_VALIDATING)) {
             return 0;
@@ -1242,29 +1389,38 @@ httpClientSideRequest(HTTPRequestPtr request)
     ObjectPtr object = request->object;
     HTTPConnectionPtr connection = request->connection;
 
+    /* NoticeError needs to be run before DiscardBody, in case
+       DiscardBody decides to close the connection. */
+
     if(urlIsLocal(object->key, object->key_size)) {
-        httpClientDiscardBody(connection);
         httpClientNoticeError(request, 503,
                               internAtom("Local PUT/POST not implemented"));
+        httpClientDiscardBody(connection);
         return 1;
     }
     if(request->from < 0 || request->to >= 0) {
-        httpClientDiscardBody(connection);
-        httpClientNoticeError(request, 503,
+        httpClientNoticeError(request, 501,
                               internAtom("Partial requests not implemented"));
+        httpClientDiscardBody(connection);
         return 1;
     }
     if(connection->reqte != TE_IDENTITY) {
-        httpClientDiscardBody(connection);
-        httpClientNoticeError(request, 411,
+        httpClientNoticeError(request, 501,
                               internAtom("Chunked requests not implemented"));
+        httpClientDiscardBody(connection);
         return 1;
     }
-
     if(connection->bodylen < 0) {
+        httpClientNoticeError(request, 502,
+                              internAtom("POST or PUT without "
+                                         "Content-Length"));
+        httpClientDiscardBody(connection);
+        return 1;
+    }
+    if(connection->reqlen < 0) {
+        httpClientNoticeError(request, 502,
+                              internAtom("Incomplete POST or PUT"));
         httpClientDiscardBody(connection);
-        httpClientNoticeError(request, 503,
-                              internAtom("Partial requests not implemented"));
         return 1;
     }
         
@@ -1278,26 +1434,31 @@ httpClientSideHandler(int status,
 {
     HTTPConnectionPtr connection = srequest->data;
     HTTPRequestPtr request = connection->request;
-    HTTPRequestPtr requestee = request->request;
-    HTTPConnectionPtr server = requestee->connection;
+    HTTPRequestPtr requestee;
+    HTTPConnectionPtr server;
     int push;
+    int code;
+    AtomPtr message = NULL;
+
+    assert(connection->flags & CONN_SIDE_READER);
 
     if((request->object->flags & OBJECT_ABORTED) || 
        !(request->object->flags & OBJECT_INPROGRESS)) {
-        httpClientDiscardBody(connection);
-        httpClientError(request, 503, internAtom("Post aborted"));
-        return 1;
+        code = request->object->code;
+        message = retainAtom(request->object->message);
+        goto fail;
     }
         
     if(status < 0) {
         do_log_error(L_ERROR, -status, "Reading from client");
-        if(status == -EDOGRACEFUL)
-            httpClientFinish(connection, 1);
-        else
-            httpClientFinish(connection, 2);
-        return 1;
+        code = 502;
+        message = internAtomError(-status, "Couldn't read from client");
+        goto fail;
     }
 
+    requestee = request->request;
+    server = requestee->connection;
+
     push = MIN(srequest->offset - connection->reqlen, 
                connection->bodylen - connection->reqoffset);
     if(push > 0) {
@@ -1307,15 +1468,31 @@ httpClientSideHandler(int status,
     }
 
     if(server->reqoffset >= connection->bodylen) {
-        connection->flags &= ~CONN_READER;
+        connection->flags &= ~(CONN_READER | CONN_SIDE_READER);
         return 1;
     }
 
     assert(status);
     do_log(L_ERROR, "Incomplete client request.\n");
-    connection->flags &= ~CONN_READER;
-    httpClientRawError(connection, 502,
-                       internAtom("Incomplete client request"), 1);
+    code = 502;
+    message = internAtom("Incomplete client request");
+
+ fail:
+    request->error_code = code;
+    if(request->error_message)
+        releaseAtom(request->error_message);
+    request->error_message = message;
+    if(request->error_headers)
+        releaseAtom(request->error_headers);
+    request->error_headers = NULL;
+
+    if(request->request) {
+        shutdown(request->request->connection->fd, 2);
+        pokeFdEvent(request->request->connection->fd, -ESHUTDOWN, POLLOUT);
+    }
+    notifyObject(request->object);
+    connection->flags &= ~CONN_SIDE_READER;
+    httpClientDiscardBody(connection);
     return 1;
 }
 
@@ -1329,8 +1506,6 @@ httpServeObject(HTTPConnectionPtr connection)
     int n, len;
     int condition_result;
 
-    assert(!(object->flags & OBJECT_VALIDATING));
-
     object->atime = current_time.tv_sec;
     objectMetadataChanged(object, 0);
 
@@ -1339,9 +1514,13 @@ httpServeObject(HTTPConnectionPtr connection)
     if((request->error_code && relaxTransparency <= 0) ||
        object->flags & OBJECT_INITIAL) {
         unlockChunk(object, i);
-        return httpClientRawError(connection,
-                                  request->error_code, 
-                                  retainAtom(request->error_message), 0);
+        if(request->error_code)
+            return httpClientRawError(connection,
+                                      request->error_code, 
+                                      retainAtom(request->error_message), 0);
+        else
+            return httpClientRawError(connection,
+                                      500, internAtom("Object vanished."), 0);
     }
 
     if(!(object->flags & OBJECT_INPROGRESS) && object->code == 0) {
@@ -1382,6 +1561,7 @@ httpServeObject(HTTPConnectionPtr connection)
         if(request->request) {
             request->request->request = NULL;
             request->request = NULL;
+            request->object->requestor = NULL;
         }
         object = makeObject(OBJECT_HTTP,
                             object->key, object->key_size, 1, 0,
@@ -1393,8 +1573,7 @@ httpServeObject(HTTPConnectionPtr connection)
         request->object = NULL;
         if(object == NULL) {
             do_log(L_ERROR, "Couldn't allocate object.");
-            return httpClientRawError(connection, 
-                                      501, 
+            return httpClientRawError(connection, 501,
                                       internAtom("Couldn't allocate object"), 
                                       1);
         }
@@ -1407,16 +1586,16 @@ httpServeObject(HTTPConnectionPtr connection)
         return httpClientNoticeRequest(request, 1);
     }
 
-    if(!(object->flags & OBJECT_INITIAL) && !object->headers) {
-        /* Aborted object */
+    if(object->flags & OBJECT_ABORTED) {
         unlockChunk(object, i);
-        return httpClientRawError(connection, object->code,
-                                  retainAtom(object->message), 0);
+        return httpClientNoticeError(request, object->code, 
+                                     retainAtom(object->message));
     }
 
     if(connection->buf == NULL)
         connection->buf = get_chunk();
     if(connection->buf == NULL) {
+        unlockChunk(object, i);
         do_log(L_ERROR, "Couldn't allocate client buffer.\n");
         connection->flags &= ~CONN_WRITER;
         httpClientFinish(connection, 1);
@@ -1508,10 +1687,10 @@ httpServeObject(HTTPConnectionPtr connection)
         }
     }
 
-    if(n >= CHUNK_SIZE - 4)
-        goto fail;
-
     n = snnprintf(connection->buf, n, CHUNK_SIZE, "\r\n\r\n");
+
+    if(n < 0)
+        goto fail;
     
     connection->offset = request->from;
 
@@ -1566,7 +1745,6 @@ delayedHttpServeObject(HTTPConnectionPtr connection)
 
     assert(connection->request->object->chunks[connection->request->from / 
                                                CHUNK_SIZE].locked > 0);
-    assert(!(connection->request->object->flags & OBJECT_VALIDATING));
 
     event = scheduleTimeEvent(-1, httpServeObjectDelayed,
                               sizeof(connection), &connection);
