@@ -592,6 +592,7 @@ typedef struct _MimeEntry {
 
 static const MimeEntryRec mimeEntries[] = {
     { "html", "text/html" },
+    { "htm", "text/html" },
     { "text", "text/plain" },
     { "txt", "text/plain" },
     { "png", "image/png" },
@@ -601,6 +602,7 @@ static const MimeEntryRec mimeEntries[] = {
     { "pdf", "image/pdf" },
     { "ps", "application/postscript" },
     { "tar", "application/x-tar" },
+    { "pac", "application/x-ns-proxy-autoconfig" },
 };
 
 static char*
@@ -635,6 +637,7 @@ localObjectMimeType(ObjectPtr object, char **encoding_return)
     for(i = 0; i < sizeof(mimeEntries) / sizeof(mimeEntries[0]); i++) {
         int len = strlen(mimeEntries[i].extension);
         if(nlen > len && 
+           name[nlen - len - 1] == '.' &&
            memcmp(name + nlen - len, mimeEntries[i].extension, len) == 0)
             return mimeEntries[i].mime;
     }
@@ -862,6 +865,8 @@ validateEntry(ObjectPtr object, int fd,
     else
         dirty = 1;
 
+    object->cache_control |= cache_control.flags;
+
     if(object->age < 0) object->age = object->date;
     if(object->age < 0) object->age = 0; /* a long time ago */
     if(object->length < 0) object->length = length;
@@ -1214,8 +1219,9 @@ destroyDiskEntry(ObjectPtr object, int d)
     } else {
         if(entry && entry->metadataDirty)
             writeoutMetadata(object);
+        makeDiskEntry(object, 1, 0);
         /* rewriteDiskEntry may change the disk entry */
-        entry = makeDiskEntry(object, 1, 0);
+        entry = object->disk_entry;
         if(entry == NULL || entry == &negativeEntry)
             return 0;
         if(entry->writeable && diskCacheWriteoutOnClose > 0)
@@ -1360,6 +1366,10 @@ objectFillFromDisk(ObjectPtr object, int offset, int chunks)
         if(object->size < o + rc)
             object->size = o + rc;
 
+        if(entry->object->length >= 0 && entry->size < 0 &&
+           entry->offset - entry->body_offset == entry->object->length)
+            entry->size = entry->object->length;
+            
         if(rc < CHUNK_SIZE - j) {
             /* Paranoia: the read may have been interrupted half-way. */
             if(entry->size < 0) {
