@@ -1,5 +1,5 @@
 /*
-Copyright (c) 2003 by Juliusz Chroboczek
+Copyright (c) 2003, 2004 by Juliusz Chroboczek
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
@@ -261,7 +261,7 @@ httpServerAbort(HTTPConnectionPtr connection, int fail,
         }
     }
     releaseAtom(message);
-    if(connection->connecting != CONNECTING_DNS)
+    if(!connection->connecting)
         httpServerFinish(connection, 1, 0);
 }
 
@@ -293,7 +293,7 @@ httpServerClientReset(HTTPRequestPtr request)
 {
     if(request->connection && 
        request->connection->fd >= 0 &&
-       request->connection->connecting != CONNECTING_DNS &&
+       !request->connection->connecting &&
        request->connection->request == request)
         pokeFdEvent(request->connection->fd, -ECLIENTRESET, POLLIN | POLLOUT);
 }
@@ -355,16 +355,13 @@ httpMakeServerRequest(char *name, int port, ObjectPtr object,
         request->request = NULL;
         requestor->request = NULL;
         object->flags &= ~(OBJECT_INPROGRESS | OBJECT_VALIDATING);
-        if(request->request) {
-            request->request->request = NULL;
-            request->request = NULL;
-        }
         releaseNotifyObject(object);
         httpDestroyRequest(request);
+        return 1;
     }
 
     if(request->wait_continue) {
-        if(server->version != HTTP_11) {
+        if(server->version == HTTP_10) {
             httpServerAbortRequest(request, 1,
                                    417, internAtom("Expectation failed"));
             return 1;
@@ -433,13 +430,15 @@ httpServerConnectionDnsHandler(int status, GethostbynameRequestPtr request)
 
     if(status <= 0) {
         AtomPtr message;
-        message = internAtomF("Host %s lookup failed: %s", 
-                              request->name->string,
+        message = internAtomF("Host %s lookup failed: %s",
+                              request->name ?
+                              request->name->string : "(unknown)",
                               request->error_message ?
                               request->error_message->string :
                               pstrerror(-status));
         do_log(L_ERROR, "Host %s lookup failed: %s (%d).\n", 
-               request->name->string,
+               request->name ?
+               request->name->string : "(unknown)",
                request->error_message ?
                request->error_message->string :
                pstrerror(-status), -status);
@@ -662,6 +661,9 @@ httpServerTrigger(HTTPServerPtr server)
 
         if(REQUEST_SIDE(server->request)) {
             rc = httpServerSideRequest(server);
+            /* If rc is 0, httpServerSideRequest didn't dequeue this
+               request.  Go through the scheduling loop again, come
+               back later. */
             if(rc <= 0) break;
             continue;
         }
@@ -816,7 +818,7 @@ httpServerSideRequest(HTTPServerPtr server)
         do_log(L_ERROR, "Couldn't write POST or PUT request.\n");
         httpServerAbortRequest(request, rc != -ECLIENTRESET, 503, 
                                internAtom("Couldn't write request"));
-        return 1;
+        return 0;
     }
     server->request = request->next;
     request->next = NULL;
@@ -887,6 +889,7 @@ httpServerDoSide(HTTPConnectionPtr connection)
                 do_log(L_ERROR, "Couldn't register object handler.\n");
             /* Fall through -- the client side will clean up. */
         }
+        client->flags |= CONN_SIDE_READER;
         do_stream(IO_READ | (done ? IO_IMMEDIATE : 0 ) | IO_NOTNOW,
                   client->fd, client->reqlen,
                   client->reqbuf, CHUNK_SIZE,
@@ -1000,8 +1003,7 @@ httpServerContinueObjectHandler(int status, ObjectHandlerPtr ohandler)
 }
 
 /* s is 0 to keep the connection alive, 1 to shutdown the connection,
-   2 to shut it down straight away, and -1 to keep the connection
-   alive and keep the current request. */
+   and -1 to keep the connection alive and keep the current request. */
 void
 httpServerFinish(HTTPConnectionPtr connection, int s, int offset)
 {
@@ -1269,6 +1271,7 @@ httpServerRestart(HTTPConnectionPtr connection)
         connection->request_last = NULL;
     }
     connection->pipelined = 0;
+    httpServerFinish(connection, 1, 0);
 }
 
 int
@@ -1288,6 +1291,8 @@ httpServerRequest(ObjectPtr object, int method, int from, int to,
     if(object->flags & OBJECT_INPROGRESS)
         return 1;
 
+    assert(requestor->request == NULL);
+
     if(requestor->requested)
         return 0;
 
@@ -1300,6 +1305,14 @@ httpServerRequest(ObjectPtr object, int method, int from, int to,
         do_log(L_FORBIDDEN, "\n");
         abortObject(object, 403, internAtom("Forbidden URL"));
         notifyObject(object);
+        if(REQUEST_SIDE(requestor)) {
+            HTTPConnectionPtr client = requestor->connection;
+            client->flags |= CONN_SIDE_READER;
+            do_stream(IO_READ | IO_IMMEDIATE | IO_NOTNOW,
+                      client->fd, client->reqlen,
+                      client->reqbuf, CHUNK_SIZE,
+                      httpClientSideHandler, client);
+        }
         return 1;
     }
 
@@ -1448,6 +1461,10 @@ httpWriteRequest(HTTPConnectionPtr connection, HTTPRequestPtr request,
             n = snnprintf(connection->reqbuf, n, CHUNK_SIZE,
                           ":%d", port);
         }
+    } else {
+        if(parentAuthCredentials)
+            n = buildServerAuthHeaders(connection->reqbuf, n, CHUNK_SIZE,
+                                       parentAuthCredentials);
     }
 
     if(bodylen >= 0)
@@ -1468,7 +1485,7 @@ httpWriteRequest(HTTPConnectionPtr connection, HTTPRequestPtr request,
         }
     }
 
-    if(method == METHOD_GET && object->etag && (from > 0 || to < 0)) {
+    if(method == METHOD_GET && object->etag && (from > 0 || to >= 0)) {
         if(request->request && request->request->request == request &&
            request->request->from == 0 && request->request->to == -1 &&
            pmmSize == 0 && pmmFirstSize == 0)
@@ -1626,8 +1643,10 @@ httpServerReplyHandler(int status,
     }
 
     if(status) {
-        if(connection->serviced >= 1)
+        if(connection->serviced >= 1) {
             httpServerRestart(connection);
+            return 1;
+        }
         if(status < 0) {
             do_log(L_ERROR, 
                    "Error reading server headers: %d\n", -status);
@@ -1660,6 +1679,7 @@ httpServerHandlerHeaders(int eof,
     CacheControlRec cache_control;
     int age = -1;
     time_t date, last_modified, expires;
+    struct timeval *init_time;
     char *etag;
     AtomPtr via, new_via;
     int expect_body;
@@ -1814,14 +1834,11 @@ httpServerHandlerHeaders(int eof,
         do_log(L_UNCACHEABLE, " (%d)\n", cache_control.flags);
     }
 
-    if(age >= 0) {
-        if(age >= current_time.tv_sec)
-            age = 0;
-        else
-            age = MIN(current_time.tv_sec - age, current_time.tv_sec);
-    } else {
-        age = current_time.tv_sec;
-    }
+    if(request->time0.tv_sec != null_time.tv_sec)
+        init_time = &request->time0;
+    else
+        init_time = &current_time;
+    age = MIN(init_time->tv_sec - age, init_time->tv_sec);
 
     if(request->method == METHOD_HEAD || 
        code < 200 || code == 204 || code == 304)
@@ -1915,9 +1932,6 @@ httpServerHandlerHeaders(int eof,
         }
         if(urlIsLocal(new_object->key, new_object->key_size))
             new_object->flags |= OBJECT_LOCAL;
-        old_object->flags |= OBJECT_SUPERSEDED;
-        old_object->flags &= ~OBJECT_INPROGRESS;
-        new_object->flags |= OBJECT_INPROGRESS;
     } else {
         new_object = object;
     }
@@ -2023,46 +2037,65 @@ httpServerHandlerHeaders(int eof,
     if(new_object->flags & OBJECT_INITIAL) {
         objectPartial(new_object, full_len, headers);
     } else {
+        if(new_object->length < 0)
+            new_object->length = full_len;
         /* XXX -- RFC 2616 13.5.3 */
         releaseAtom(headers);
     }
 
-    new_object->flags &= ~OBJECT_VALIDATING;
-    notifyObject(new_object);
-
     if(supersede) {
         assert(new_object != old_object);
         supersedeObject(old_object);
     }
+
     if(new_object != old_object) {
         old_object->flags &= ~OBJECT_VALIDATING;
+        new_object->flags |= OBJECT_INPROGRESS;
+        /* Signal the client side to switch to the new object -- see
+           httpClientGetHandler.  If it doesn't, we'll give up on this
+           request below. */
+        new_object->flags |= OBJECT_MUTATING;
+        request->can_mutate = new_object;
+        notifyObject(old_object);
+        request->can_mutate = NULL;
+        new_object->flags &= ~OBJECT_MUTATING;
+        old_object->flags &= ~OBJECT_INPROGRESS;
+        if(request->object != new_object) {
+            if(request->request)
+                request->request->request = NULL;
+            request->request = NULL;
+            request->object = new_object;
+        }
         releaseNotifyObject(old_object);
         old_object = NULL;
-        request->object = new_object;
         object = new_object;
     } else {
         objectMetadataChanged(new_object, 0);
     }
 
-    if(!expect_body) {
-        if(request->method == METHOD_GET && object->length != 0) {
-            do_log(L_ERROR, "No content received.\n");
-            httpServerAbort(connection, 1, 502, 
-                            internAtom("No content received"));
-            return 1;
-        } else {
-            if(code == 412) {
-                /* 412 replies contain a useless body.  For now, we
-                   drop the connection. */
-                httpServerFinish(connection, 1, 0);
-            } else {
-                httpServerFinish(connection, 0, rc);
-            }
-            notifyObject(object);
-            return 1;
-        }
+    if(object->flags & OBJECT_VALIDATING) {
+        object->flags &= ~OBJECT_VALIDATING;
+        notifyObject(object);
     }
 
+    if(!expect_body) {
+        httpServerFinish(connection, 0, rc);
+        return 1;
+    }
+
+    if(request->request == NULL) {
+        httpServerFinish(connection, 1, 0);
+        return 1;
+    }
+
+    if(code == 412) {
+        /* 412 replies contain a useless body.  For now, we
+           drop the connection. */
+        httpServerFinish(connection, 1, 0);
+        return 1;
+    }
+
+
     if(request->persistent) {
         if(request->method != METHOD_HEAD && 
            connection->te == TE_IDENTITY && len < 0) {
@@ -2116,7 +2149,6 @@ httpServerHandlerHeaders(int eof,
                     }
                 }
                 httpServerFinish(connection, 0, 0);
-                notifyObject(object);
                 return 1;
             }
         }
@@ -2133,7 +2165,7 @@ httpServerHandlerHeaders(int eof,
                             internAtom("Server closed connection"));
             return 1;
         } else {
-            if(code != 206 && object->length < 0) {
+            if(code != 206 && eof != -ECLIENTRESET && object->length < 0) {
                 object->length = object->size;
                 objectMetadataChanged(object, 0);
             }
@@ -2208,7 +2240,7 @@ httpServerIndirectHandlerCommon(HTTPConnectionPtr connection, int eof)
                             internAtom("Server dropped connection"));
             return 1;
         } else {
-            if(request->object->length < 0 &&
+            if(request->object->length < 0 && eof != -ECLIENTRESET &&
                (request->to < 0 || request->to > request->object->size)) {
                 request->object->length = request->object->size;
                 objectMetadataChanged(request->object, 0);
@@ -2250,6 +2282,11 @@ httpServerReadData(HTTPConnectionPtr connection, int immediate)
     ObjectPtr object = request->object;
     int to = -1;
 
+    if(request->request == NULL) {
+        httpServerFinish(connection, 1, 0);
+        return 1;
+    }
+
     if(request->to >= 0)
         to = request->to;
     else
